from abc import ABC, abstractmethod

import yaml
from delta.tables import DeltaTable
from pyspark.sql import DataFrame

from ingestion_lib.utils.data_contract import TableContract


def get_row_count_written(df: DataFrame, location: str = None, table_name: str = None) -> int:
    """
    Unchecked
    :param df:
    :param location:
    :param table_name:
    :return:
    """
    if not (location or table_name):
        raise ValueError("Location or table_name must be provided.")
    if location and table_name:
        raise ValueError("Only one of the location or table_name arguments must be provided.")
    spark = df.sparkSession
    if location:
        delta_table = DeltaTable.forPath(spark, location)
    else:
        delta_table = DeltaTable.forName(spark, table_name)

    # Get the latest 5 operations from the history of the Delta Table
    history = delta_table.history(5).collect()
    if not history:
        # cannot use logger here due to circular dependency
        print(f"Cannot get written row count. History is empty. Table: {table_name}, Location: {location}")
        return -1
    for history_record in history:  # History is in descending order by default
        if (
                "operation" in history_record
                and (history_record["operation"] == "WRITE" or history_record["operation"] == "CREATE TABLE AS SELECT")
                and "operationMetrics" in history_record
                and "numOutputRows" in history_record["operationMetrics"]
        ):
            try:
                return int(history_record["operationMetrics"]["numOutputRows"])
            except (ValueError, TypeError):
                print(
                    f"Cannot get written row count. Number of output rows is not an integer. "
                    f"Table: {table_name}, Location: {location}")
                return -1
    print(
        f"Cannot get written row count. Cannot find number of output rows in history. "
        f"Table: {table_name}, Location: {location}")
    return -1


def get_delta_write_options(table_contract: TableContract) -> dict:
    """
    Returns dictionary with replaceWhere condition if watermark column exists and full load is set to false.
    Returns empty dictionary if there is no watermark column mentioned or full load is set to true.
    """
    if not table_contract.watermark_columns or table_contract.full_load or table_contract.load_type == "one_time":
        return {"mergeSchema": True}
    elif len(table_contract.watermark_columns) == 1:
        return {
            "replaceWhere":
                f"{table_contract.watermark_columns[0]} >= '{table_contract.lower_bound}' "
                f"AND {table_contract.watermark_columns[0]} <= '{table_contract.upper_bound}'",
            "mergeSchema": True,
        }
    else:
        watermark_columnss = ", ".join(table_contract.watermark_columns)
        return {
            "replaceWhere":
                f"greatest({watermark_columnss}) >= '{table_contract.lower_bound}' "
                f"AND greatest({watermark_columnss}) <= '{table_contract.upper_bound}'",
            "mergeSchema": True,
        }


class WorkflowGenerator(ABC):
    def __init__(self, ingestion_contract_path, workflow_template_path, output_path):
        self.ingestion_contract_path = ingestion_contract_path
        self.workflow_template_path = workflow_template_path
        self.output_path = output_path

    def read_yaml(self, file_path):
        with open(file_path, 'r') as file:
            return yaml.safe_load(file)

    def write_yaml(self, data, file_path):
        with open(file_path, 'w') as file:
            yaml.safe_dump(data, file, default_flow_style=False, sort_keys=False)

    @abstractmethod
    def generate_workflow(self):
        pass


class DatabricksWorkflowGenerator(WorkflowGenerator):
    def __init__(self, ingestion_contract_path, workflow_template_path, output_path, task_config):
        super().__init__(ingestion_contract_path, workflow_template_path, output_path)
        self.task_config = task_config

    def transform_dataset_name(self, name):
        return name.replace('.', '_')

    def generate_workflow(self):
        ingestion_contract = self.read_yaml(self.ingestion_contract_path)
        datasets = ingestion_contract.get('datasets', [])

        workflow_template = self.read_yaml(self.workflow_template_path)
        tasks_section = workflow_template.setdefault('resources', {}).setdefault('jobs', {}).setdefault('ingestion', {}).setdefault('tasks', [])

        existing_tasks = {task['task_key']: task for task in tasks_section if 'task_key' in task}

        for dataset in datasets:
            task_key = self.transform_dataset_name(dataset['name'])
            task_data = {'task_key': task_key}
            task_data.update(self.task_config)  # Merge the task configuration

            if task_key in existing_tasks:
                existing_tasks[task_key].update(task_data)
            else:
                existing_tasks[task_key] = task_data

        if existing_tasks:
            workflow_template['resources']['jobs']['ingestion']['tasks'] = list(existing_tasks.values())
        else:
            workflow_template['resources']['jobs']['ingestion'].pop('tasks', None)

        self.write_yaml(workflow_template, self.output_path)




from pyspark.sql import SparkSession

from ingestion_lib.extractors.base import Extractor
from ingestion_lib.utils.data_contract import DbCredentials, TableContract


class OracleExtractor(Extractor):

    def __init__(self, table_contract: TableContract, spark: SparkSession):
        super().__init__(table_contract, spark)

    def extract_data(self):
        return super().extract_data()

    def creds(self) -> DbCredentials:
        return self.table_contract.credentials

    def load_data_query(self, query: str):
        """
        unchecked
        :param query:
        :return:
        """
        oracle_query = f"({query}) temp"
        print(oracle_query)
        return (
            self.spark.read.format("jdbc")
            .option("url", self.creds().jdbc_url)
            .option("user", self.creds().user)
            .option("password", self.creds().password)
            .option("driver", "oracle.jdbc.driver.OracleDriver")
            .option("dbtable", oracle_query)
            .load()
        )






from typing import Optional, List

from pydantic import BaseModel


class DbCredentials(BaseModel):
    """
    Class containing credentials like user, password and jdbc_url for connecting to oracle_EBS
    """
    user: str
    password: str
    jdbc_url: str


class DataContract(BaseModel):
    batch_timestamp: str
    scope: str
    env_key: str
    target_schema: str
    table_name: str


class TableContract(DataContract):
    table_name: str
    schema_name: str

    watermark_columns: Optional[List[str]]
    lower_bound: Optional[str] = None
    upper_bound: Optional[str] = None
    full_load: bool = False
    load_type: str = 'incremental'
    target_schema: str
    credentials: Optional[DbCredentials] = None

    class Config:
        anystr_lower = True
        allow_population_by_field_name = True


class APIDataContract(DataContract):
    base_url: str
    api_key: str  # or other authentication fields
    spec_path: str
    endpoint_url: str
    model: str # tricky part to be honest
    credentials: DbCredentials



from abc import ABC, abstractmethod
from typing import Union

from pyspark.sql.session import SparkSession, DataFrame

from ingestion_lib.utils.data_contract import TableContract, APIDataContract


class Extractor(ABC):

    def __init__(self, contract: Union[TableContract|APIDataContract], spark: SparkSession):
        self.data_contract = contract
        self.spark = spark

    @abstractmethod
    def extract_data(self):
        pass


class JdbcExtractor(ABC):
    def __init__(self, table_contract: TableContract, spark: SparkSession):
        self.table_contract = table_contract
        self.spark = spark


    @abstractmethod
    def creds(self):
        # TODO: I'm not sure about this being abstract. Probably better to parse proper section from contract
        pass

    @abstractmethod
    def load_data_query(self, query: str):
        pass

    def extract_data(self) -> DataFrame:
        """
        :return:
        """
        # TODO: Add invalid type checks
        select_query = self.__build_select_query()
        condition = self.__build_condition()
        query = f"{select_query}{condition}"
        data = self.load_data_query(query)
        # TODO: Add logging at debug level
        if self.table_contract.watermark_columns and len(self.table_contract.watermark_columns) > 1:
            # dropping column 'watermark_column' which is having max timestamp if there are multiple timestamp columns
            data = data.drop("_watermark_column_")
        return data

    def __build_condition(self) -> str:
        """
        **Step 5: Build condition (if watermark columns are present)**
        Description: If watermark columns are present in the `self.table_contract` object, the method builds a condition clause using the `__build_condition` method.
        Details: The condition clause is constructed based on the watermark columns, and is used to filter the data retrieved from the SQL Server database.
        :return:
        """

        if not self.table_contract.watermark_columns or self.table_contract.full_load or self.table_contract.load_type == "one_time":
            return ""
        elif len(self.table_contract.watermark_columns) == 1:
            return (
                    f" WHERE {self.table_contract.watermark_columns[0]} >= '{self.table_contract.lower_bound}' "
                    + f"AND {self.table_contract.watermark_columns[0]} < '{self.table_contract.upper_bound}'"
            )
        else:
            return (
                    f" WHERE _watermark_column_ >= '{self.table_contract.lower_bound}' " + f"AND _watermark_column_ < '{self.table_contract.upper_bound}'"
            )

    def __build_select_query(self) -> str:
        """
        **Step 4: Build select query (if no invalid types)**
        Description: If no invalid types are present, the method builds a select query using the `__build_select_query` method.
        Details: The select query is constructed based on the table and schema information, and is used to retrieve data from the SQL Server database.

        :return:
        """
        table = self.table_contract.table_name
        schema = self.table_contract.schema_name
        if not self.table_contract.watermark_columns or self.table_contract.full_load == "true" or len(
                self.table_contract.watermark_columns) == 1:
            return f"SELECT * FROM {schema}.{table}"
        else:
            watermark_columns = ", ".join([f"({col})" for col in self.table_contract.watermark_columns])
            return f"""
                    SELECT *
                    FROM (
                        SELECT 
                        *, 
                        (SELECT MAX(_watermark_column_)
                            FROM (VALUES {watermark_columns}) AS last_updated(_watermark_column_)) 
                        AS _watermark_column_
                    
                    """





from pyspark.sql.types import StructType, StructField, StringType

HOLMAN_VEHICLES_SCHEMA = StructType([
    StructField("lesseeCode", StringType(), True),
    StructField("ariVehicleNumber", StringType(), True),
    StructField("clientVehicleNumber", StringType(), True),
    StructField("vin", StringType(), True),
    StructField("status", StringType(), True),
    StructField("assignedStatus", StringType(), True),
    StructField("division", StringType(), True),
    StructField("prefix", StringType(), True),
    StructField("clientData1", StringType(), True),
    StructField("clientData2", StringType(), True),
    StructField("clientData3", StringType(), True),
    StructField("clientData4", StringType(), True),
    StructField("clientData5", StringType(), True),
    StructField("clientData6", StringType(), True),
    StructField("clientData7", StringType(), True),
    StructField("driverClass", StringType(), True),
    StructField("exec", StringType(), True),
    StructField("auxData1", StringType(), True),
    StructField("auxData2", StringType(), True),
    StructField("auxData3", StringType(), True),
    StructField("auxData4", StringType(), True),
    StructField("auxData5", StringType(), True),
    StructField("auxData6", StringType(), True),
    StructField("auxData7", StringType(), True),
    StructField("auxData8", StringType(), True),
    StructField("auxData9", StringType(), True),
    StructField("auxData10", StringType(), True),
    StructField("auxData11", StringType(), True),
    StructField("auxData12", StringType(), True),
    StructField("auxData13", StringType(), True),
    StructField("auxData14", StringType(), True),
    StructField("auxDate1", StringType(), True),
    StructField("auxDate2", StringType(), True),
    StructField("auxDate3", StringType(), True),
    StructField("auxDate4", StringType(), True),
    StructField("firstName", StringType(), True),
    StructField("lastName", StringType(), True),
    StructField("email", StringType(), True),
    StructField("cellPhone", StringType(), True),
    StructField("workPhone", StringType(), True),
    StructField("workPhoneExtension", StringType(), True),
    StructField("homePhone", StringType(), True),
    StructField("addressLine1", StringType(), True),
    StructField("addressLine2", StringType(), True),
    StructField("addressLine3", StringType(), True),
    StructField("city", StringType(), True),
    StructField("stateProvince", StringType(), True),
    StructField("zipPostalCode", StringType(), True),
    StructField("modelYear", StringType(), True),
    StructField("makeVin", StringType(), True),
    StructField("makeClient", StringType(), True),
    StructField("modelVin", StringType(), True),
    StructField("modelClient", StringType(), True),
    StructField("engineType", StringType(), True),
    StructField("driveline", StringType(), True),
    StructField("series", StringType(), True),
    StructField("fuelType", StringType(), True),
    StructField("fuelTypeDescription", StringType(), True),
    StructField("fuelTypeCode", StringType(), True),
    StructField("fuelCapacity", StringType(), True),
    StructField("fuelMeasureType", StringType(), True),
    StructField("odometer", StringType(), True),
    StructField("odometerDate", StringType(), True),
    StructField("hourMeter", StringType(), True),
    StructField("hourMeterDate", StringType(), True),
    StructField("licensePlate", StringType(), True),
    StructField("tagStateProvince", StringType(), True),
    StructField("plateType", StringType(), True),
    StructField("renewalDate", StringType(), True),
    StructField("assetType", StringType(), True),
    StructField("assetSubType", StringType(), True),
    StructField("orderType", StringType(), True),
    StructField("deliveryDate", StringType(), True),
    StructField("onRoadDate", StringType(), True),
    StructField("outOfServiceDate", StringType(), True),
    StructField("soldDate", StringType(), True),
    StructField("soldAmount", StringType(), True),
    StructField("saleOdometer", StringType(), True),
    StructField("capCost", StringType(), True),
    StructField("gvwr", StringType(), True),
    StructField("gcwr", StringType(), True),
    StructField("curbWeight", StringType(), True),
    StructField("registeredVehicleWeight", StringType(), True),
    StructField("vendor", StringType(), True),
    StructField("leaseTerm", StringType(), True),
    StructField("monthsBilled", StringType(), True),
    StructField("monthsInService", StringType(), True),
    StructField("remainingBookValue", StringType(), True),
    StructField("telematicsDeviceId", StringType(), True),
    StructField("telematicsDeviceModel", StringType(), True),
    StructField("telematicsDeviceVendor", StringType(), True),
    StructField("orderReceivedDate", StringType(), True),
    StructField("orderPlacedDate", StringType(), True),
    StructField("scheduledProductionDate", StringType(), True),
    StructField("deliveryPaperMailed", StringType(), True),
    StructField("shipDate", StringType(), True),
    StructField("deliveredToDealerDate", StringType(), True),
    StructField("notifiedOfDelivery", StringType(), True),
    StructField("upfitInvoiceDate", StringType(), True),
    StructField("titleLocation", StringType(), True),
    StructField("titleOwnerLessor", StringType(), True)
])



from ingestion_lib.model.holman import HOLMAN_VEHICLES_SCHEMA


class SchemaCreator:
    def create_schema(self):
        raise NotImplementedError("Subclasses must implement create_schema method")


class VehicleSchemaCreator(SchemaCreator):
    def create_schema(self):
        return HOLMAN_VEHICLES_SCHEMA


class SchemaFactory:
    @staticmethod
    def get_schema(property_value):
        if property_value == "vehicles":
            return VehicleSchemaCreator().create_schema()
        # Add more conditions for different schema types
        else:
            raise ValueError(f"Unknown property value: {property_value}")



from pyspark.sql.functions import lit
from ingestion_lib.extractors.base import Extractor
from pyspark.sql import DataFrame, SparkSession

from ingestion_lib.extractors.oracle import OracleExtractor
from ingestion_lib.utils.data_contract import TableContract
from ingestion_lib.utils.databricks_utils import get_row_count_written, get_delta_write_options


class DatabricksIngestionProvider:
    def __init__(self, extractor: Extractor):
        self.extractor = extractor

    @staticmethod
    def init_oracle_with_table_contract(table_contract: TableContract, spark: SparkSession):
        # TODO: create better logic to handle Extractor Creation
        return DatabricksIngestionProvider(OracleExtractor(table_contract, spark))

    def normalize_data(self, data) -> DataFrame:
        for column in data.columns:
            data = data.withColumnRenamed(column, column.replace(" ", "_"))
        return data

    def write_data(self, data, target_schema, mount_point, table_name, options):
        # Logic to write data to Databricks Delta Lake
        location = f"{mount_point}/{target_schema}/{table_name}"
        (data.distinct().write.format("delta").mode("overwrite").options(**options).save(location))
        return get_row_count_written(data, location=location)

    def write_delta_unity(self, data, target_schema, table_name, options):
        full_name = f"{target_schema}.{table_name}"
        print(f"Saving into {full_name}")
        print(f"Options: {options}")
        data.printSchema()
        data.show()
        (data.distinct().write.format("delta").mode("overwrite").options(**options).saveAsTable(full_name))
        return get_row_count_written(data, table_name=table_name)

    def execute_ingestion(self):
        """
        Missing steps:
        **Step 9: Run data quality checks (optional)**
        Description: Run data quality checks on the DataFrame if a contract is provided.
        Details: If a contract is provided and it contains "dq_rules", the `DataQualityChecker` class is used to run data quality checks.

        **Step 11: Log row count and return**
        Description: Log the row count of the written DataFrame and return.
        Details: The row count is logged, and the method returns the row count.
        :return:
        """
        table_contract = self.extractor.table_contract
        data = self.extractor.extract_data()
        normalized_data = self.__add_timestamp_column(table_contract.batch_timestamp, self.normalize_data(data))

        result = self.write_delta_unity(normalized_data, table_contract.target_schema,
                                        table_contract.table_name, options=get_delta_write_options(table_contract))

        print(f"Table {table_contract.table_name} is successfully ingested with the row count {result}")


    def __add_timestamp_column(self, batch_timestamp: str, df: DataFrame) -> DataFrame:
        """
        Adds timestamp column to the input DataFrame.

        Parameters:
        - load_timestamp: The timestamp to be added to the DataFrame.
        - df: The Spark DataFrame to which the timestamp column will be added.

        Returns:
        - The Spark DataFrame with a new column named 'load_timestamp' that contains the provided timestamp.
        """
        return df.withColumn("load_timestamp", lit(batch_timestamp).cast("timestamp"))



from abc import abstractmethod
from typing import Union

import httpx
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, StructField, StringType

from ingestion_lib.extractors.base import Extractor
from ingestion_lib.model.base import SchemaFactory
from ingestion_lib.model.holman import HOLMAN_VEHICLES_SCHEMA
from ingestion_lib.utils.data_contract import TableContract, APIDataContract


class RestAPIExtractor(Extractor):

    def __init__(self, contract: APIDataContract, spark: SparkSession):
        super().__init__(contract, spark)


    def extract_data(self):
        client = self.init_client()
        header = self.auth_header(client)
        collection = self.get_collection(header, client)
        return self.convert_collection(collection)

    def init_client(self):
        return httpx.Client(base_url=self.data_contract.base_url)

    @abstractmethod
    def auth_header(self, client):
        pass

    @abstractmethod
    def get_collection(self, header, client):
        pass

    def convert_collection(self, collection):
        return collection

class HolmanRestAPIExtractor(RestAPIExtractor):

    def get_collection(self, header, client):
        url = f"{self.data_contract.base_url}/{self.data_contract.endpoint_url}"
        model_name = self.data_contract.model
        response = client.get(url, headers=header)
        initial_data = response.json()
        result = initial_data[model_name]

        total_pages = int(initial_data.get('totalPages', 1))

        for page in range(2, total_pages + 1):
            response =client.get(f"{url}?pageNumber={page}", headers=header)
            response.raise_for_status()
            page_data = response.json()
            result.extend(page_data[model_name])
        return result

    def auth_header(self, client):
        """
        Authenticate with the API and return the token.
        """
        url = f"{self.data_contract.base_url}/v1/users/authenticate"
        headers = {"Content-Type": "application/json"}
        payload = {
            "userName": self.data_contract.credentials.user,
            "password": self.data_contract.credentials.password
        }

        response = client.post(url=url, json=payload, headers=headers)
        if response.status_code == 200:
            print("Authentication successful")
            return {'Authorization': f'Bearer {response.json().get("token")}'}
        else:
            print("Failed to authenticate")
            print("Status code:", response.status_code)
            print("Response:", response.text)
        return None

    def convert_collection(self, collection):
        schema = SchemaFactory.get_schema(self.data_contract.table_name)
        df = self.spark.createDataFrame(collection, schema=schema)
        return df

    def __init__(self, contract: APIDataContract, spark: SparkSession):
        super().__init__(contract, spark)
